package ru.dbotthepony.mc.otm.datagen.loot

import com.mojang.datafixers.util.Pair
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import net.minecraft.data.DataGenerator
import net.minecraft.data.loot.LootTableProvider
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.level.ItemLike
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.storage.loot.LootPool
import net.minecraft.world.level.storage.loot.LootTable
import net.minecraft.world.level.storage.loot.ValidationContext
import net.minecraft.world.level.storage.loot.entries.LootItem
import net.minecraft.world.level.storage.loot.functions.CopyNbtFunction
import net.minecraft.world.level.storage.loot.parameters.LootContextParamSet
import net.minecraft.world.level.storage.loot.parameters.LootContextParamSets
import net.minecraft.world.level.storage.loot.providers.nbt.ContextNbtProvider
import ru.dbotthepony.mc.otm.block.entity.MatteryBlockEntity
import ru.dbotthepony.mc.otm.block.entity.MatteryWorkerBlockEntity
import java.util.function.BiConsumer
import java.util.function.Consumer
import java.util.function.Supplier

private typealias LootBuilderCallback = BiConsumer<ResourceLocation, LootTable.Builder>
private typealias LootBuilderWorker = Consumer<LootBuilderCallback>
private typealias LootBuilderWorkerSupplier = Supplier<LootBuilderWorker>
private typealias LootTuple = Pair<LootBuilderWorkerSupplier, LootContextParamSet>

fun singleLootPool(f: (LootPool.Builder) -> Unit): LootTable.Builder {
	return LootTable.lootTable().withPool(LootPool.lootPool().also(f))
}

data class NbtCopy(val source: String, val destination: String, val strategy: CopyNbtFunction.MergeStrategy = CopyNbtFunction.MergeStrategy.REPLACE)

fun TileNbtCopy(source: String, strategy: CopyNbtFunction.MergeStrategy = CopyNbtFunction.MergeStrategy.REPLACE): NbtCopy {
	return NbtCopy(source, "BlockEntityTag.$source", strategy)
}

private val basicTags = arrayOf(
	TileNbtCopy(MatteryBlockEntity.REDSTONE_SIGNAL_KEY),
	TileNbtCopy(MatteryBlockEntity.REDSTONE_SETTING_KEY),
)

private val poweredTags = arrayOf(
	*basicTags,
	TileNbtCopy(MatteryBlockEntity.ENERGY_KEY),
	TileNbtCopy(MatteryBlockEntity.BATTERY_KEY),
)

private val workerTags = arrayOf(
	*poweredTags,
	TileNbtCopy(MatteryWorkerBlockEntity.JOB_KEY),
	TileNbtCopy(MatteryWorkerBlockEntity.WORK_TICKS_KEY),
)

private val poweredMatterWorker = arrayOf(
	*workerTags,
	TileNbtCopy(MatteryBlockEntity.MATTER_STORAGE_KEY),
)

class LootTables(generator: DataGenerator) : LootTableProvider(generator) {
	private val providers = ArrayList<LootTuple>()

	override fun getTables(): List<LootTuple> {
		return providers
	}

	override fun validate(map: MutableMap<ResourceLocation, LootTable>, validationtracker: ValidationContext) {}

	fun block(provider: LootBuilderWorker) {
		providers.add(Pair(Supplier {provider}, LootContextParamSets.BLOCK))
	}

	private val seenStuff = ObjectArraySet<Any>()

	fun dropsSelf(block: Block) {
		if (!seenStuff.add(block)) {
			throw IllegalStateException("Already seen $block")
		}

		block {
			it.accept(block.lootTable, singleLootPool {
				it.add(LootItem.lootTableItem(block))
			})
		}
	}

	fun dropsOther(block: Block, other: ItemLike) {
		if (!seenStuff.add(block)) {
			throw IllegalStateException("Already seen $block")
		}

		block {
			it.accept(block.lootTable, singleLootPool {
				it.add(LootItem.lootTableItem(other))
			})
		}
	}

	fun dropsSelf(vararg blocks: Block) {
		blocks.forEach(this::dropsSelf)
	}

	fun dropsSelf(blocks: Collection<Block>) {
		blocks.forEach(this::dropsSelf)
	}

	fun tile(block: Block, f: (CopyNbtFunction.Builder) -> Unit = {}) {
		if (!seenStuff.add(block)) {
			throw IllegalStateException("Already seen $block")
		}

		block {
			it.accept(block.lootTable, singleLootPool {
				it.add(LootItem.lootTableItem(block).also {
					it.apply(CopyNbtFunction.copyData(ContextNbtProvider.BLOCK_ENTITY).also {
						it.copy("Name", "BlockEntityTag.Name")
						f(it)
					})
				})
			})
		}
	}

	fun tile(block: Block, vararg tags: NbtCopy) {
		tile(block) {
			for ((source, destination, strategy) in tags) {
				it.copy(source, destination, strategy)
			}
		}
	}

	fun basicTile(block: Block, vararg tags: NbtCopy) {
		tile(block) {
			for ((source, destination, strategy) in tags) {
				it.copy(source, destination, strategy)
			}

			for ((source, destination, strategy) in basicTags) {
				it.copy(source, destination, strategy)
			}
		}
	}

	fun poweredTile(block: Block, vararg tags: NbtCopy) {
		tile(block) {
			for ((source, destination, strategy) in tags) {
				it.copy(source, destination, strategy)
			}

			for ((source, destination, strategy) in poweredTags) {
				it.copy(source, destination, strategy)
			}
		}
	}

	fun workerTile(block: Block, vararg tags: NbtCopy) {
		tile(block) {
			for ((source, destination, strategy) in tags) {
				it.copy(source, destination, strategy)
			}

			for ((source, destination, strategy) in workerTags) {
				it.copy(source, destination, strategy)
			}
		}
	}

	fun matterWorkerTile(block: Block, vararg tags: NbtCopy) {
		tile(block) {
			for ((source, destination, strategy) in tags) {
				it.copy(source, destination, strategy)
			}

			for ((source, destination, strategy) in poweredMatterWorker) {
				it.copy(source, destination, strategy)
			}
		}
	}

	fun tile(block: Block, vararg tags: String) {
		tile(block, *tags.map { NbtCopy(it, "BlockEntityTag.$it", CopyNbtFunction.MergeStrategy.REPLACE) }.toTypedArray())
	}

	// fix overload resolution by adding extra required argument
	fun basicTile(block: Block, f: String, vararg tags: String) {
		basicTile(block, *tags.map { TileNbtCopy(it) }.toMutableList().also { it.add(TileNbtCopy(f)) }.toTypedArray())
	}

	fun poweredTile(block: Block, f: String, vararg tags: String) {
		poweredTile(block, *tags.map { TileNbtCopy(it) }.toMutableList().also { it.add(TileNbtCopy(f)) }.toTypedArray())
	}

	fun workerTile(block: Block, f: String, vararg tags: String) {
		workerTile(block, *tags.map { TileNbtCopy(it) }.toMutableList().also { it.add(TileNbtCopy(f)) }.toTypedArray())
	}

	fun matterWorkerTile(block: Block, f: String, vararg tags: String) {
		matterWorkerTile(block, *tags.map { TileNbtCopy(it) }.toMutableList().also { it.add(TileNbtCopy(f)) }.toTypedArray())
	}
}
