package ru.dbotthepony.mc.otm.container

import net.minecraft.world.item.ItemStack
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.ListTag
import net.minecraft.nbt.Tag
import net.minecraft.world.Container
import kotlin.jvm.JvmOverloads
import net.minecraft.world.entity.player.Player
import net.minecraft.world.level.block.entity.BlockEntity
import ru.dbotthepony.mc.otm.core.ifHas
import java.util.*

@Suppress("UNUSED")
open class MatteryContainer(val watcher: Runnable, private val size: Int) : Container, Iterable<ItemStack> {
	constructor(watcher: BlockEntity, size: Int) : this(watcher::setChanged, size)
	constructor(size: Int) : this({}, size)

	init {
		require(size >= 0) { "Invalid container size $size" }
	}

	private var ignoreChangeNotifications = 0
	protected val slots: Array<ItemStack> = Array(size) { ItemStack.EMPTY }
	private val trackedSlots: Array<ItemStack> = Array(size) { ItemStack.EMPTY }

	final override fun getContainerSize() = size

	fun startIgnoreUpdates() {
		if (++ignoreChangeNotifications == 1) {
			startedIgnoringUpdates()
		}
	}

	fun stopIgnoreUpdates() {
		if (--ignoreChangeNotifications == 0) {
			stoppedIgnoringUpdates()
		}
	}

	protected open fun startedIgnoringUpdates() {}
	protected open fun stoppedIgnoringUpdates() {}

	private fun deserializeNBT(tag: CompoundTag?) {
		Arrays.fill(slots, ItemStack.EMPTY)

		if (tag == null) {
			setChanged()
			return
		}

		// нам не интересен размер
		tag.ifHas("items", ListTag::class.java) {
			for (i in 0 until it.size.coerceAtMost(size)) {
				slots[i] = ItemStack.of(it[i] as CompoundTag)
			}
		}

		setChanged()
	}

	fun deserializeNBT(tag: ListTag?) {
		Arrays.fill(slots, ItemStack.EMPTY)

		if (tag == null) {
			setChanged()
			return
		}

		for (i in 0 until tag.size.coerceAtMost(size)) {
			slots[i] = ItemStack.of(tag[i] as CompoundTag)
		}

		setChanged()
	}

	fun deserializeNBT(tag: Tag?) {
		when (tag) {
			is CompoundTag -> deserializeNBT(tag)
			is ListTag -> deserializeNBT(tag)
			else -> {
				Arrays.fill(slots, ItemStack.EMPTY)
				setChanged()
			}
		}
	}

	open fun setChanged(slot: Int, new: ItemStack, old: ItemStack) {
		if (ignoreChangeNotifications == 0) watcher.run()
	}

	fun serializeNBT(): ListTag {
		return ListTag().also {
			for (i in 0 until size) {
				it.add(this[i].serializeNBT())
			}
		}
	}

	fun hasEmptySlot(): Boolean {
		for (i in 0 until size) {
			if (this[i].isEmpty) {
				return true
			}
		}

		return false
	}

	fun handler(
		insert_validator: (slot: Int, stack: ItemStack) -> Boolean,
		extract_validator: (slot: Int, amount: Int, stack: ItemStack) -> Boolean
	): MatteryContainerHandler {
		return MatteryContainerHandler(this, object : MatteryContainerFilter {
			override fun canInsert(slot: Int, stack: ItemStack) = insert_validator(slot, stack)
			override fun canExtract(slot: Int, amount: Int, stack: ItemStack) = extract_validator(slot, amount, stack)
		})
	}

	fun handler(
		filter: MatteryContainerFilter
	): MatteryContainerHandler {
		return MatteryContainerHandler(this, filter)
	}

	fun handler(insert_validator: (Int, ItemStack) -> Boolean): MatteryContainerHandler {
		return MatteryContainerHandler(this, object : MatteryContainerFilter {
			override fun canInsert(slot: Int, stack: ItemStack) = insert_validator(slot, stack)
			override fun canExtract(slot: Int, amount: Int, stack: ItemStack) = false
		})
	}

	fun handler(): MatteryContainerHandler {
		return MatteryContainerHandler(this)
	}

	open fun getMaxStackSize(slot: Int) = maxStackSize

	open fun getMaxStackSizeWithItem(slot: Int): Int {
		val item = this[slot]

		if (!item.isEmpty) {
			return getMaxStackSize(slot).coerceAtMost(item.maxStackSize)
		}

		return getMaxStackSize(slot)
	}

	@JvmOverloads
	fun addItem(stack: ItemStack, range: IntRange, simulate: Boolean = false): ItemStack {
		if (range.last >= size || range.first < 0)
			throw IllegalArgumentException("Invalid range: $range")

		if (stack.isEmpty)
			return stack

		val copy = stack.copy()

		// двигаем в одинаковые слоты
		for (slot in range) {
			if (ItemStack.isSameItemSameTags(slots[slot], copy)) {
				val slotStack = slots[slot]
				val slotLimit = getMaxStackSize(slot).coerceAtMost(slotStack.maxStackSize)

				if (slotStack.count < slotLimit) {
					val newCount = (slotStack.count + copy.count).coerceAtMost(slotLimit)
					val diff = newCount - slotStack.count

					if (!simulate) {
						val old = slotStack.copy()
						slotStack.count = newCount
						trackedSlots[slot] = slotStack.copy()
						setChanged(slot, slotStack, old)
					}

					copy.shrink(diff)

					if (copy.isEmpty) {
						return copy
					}
				}
			}
		}

		// двигаем в пустые слоты
		for (slot in range) {
			if (slots[slot].isEmpty) {
				val diff = copy.count.coerceAtMost(getMaxStackSize(slot).coerceAtMost(copy.maxStackSize))

				if (!simulate) {
					val copyToPut = copy.copy()
					copyToPut.count = diff
					this[slot] = copyToPut
				}

				copy.shrink(diff)

				if (copy.isEmpty) {
					return copy
				}
			}
		}

		return copy
	}

	fun addItem(stack: ItemStack, simulate: Boolean): ItemStack {
		return addItem(stack, 0 until size, simulate)
	}

	@JvmOverloads
	fun fullyAddItem(stack: ItemStack, start: Int = 0, end: Int = size - 1): Boolean {
		return fullyAddItem(stack, start .. end)
	}

	fun fullyAddItem(stack: ItemStack, range: IntRange): Boolean {
		if (!addItem(stack, range, true).isEmpty)
			return false

		return addItem(stack, range, false).isEmpty
	}

	override fun isEmpty(): Boolean {
		for (stack in slots)
			if (!stack.isEmpty)
				return false

		return true
	}

	operator fun get(slot: Int) = getItem(slot)
	operator fun set(slot: Int, stack: ItemStack) = setItem(slot, stack)

	operator fun contains(other: ItemStack): Boolean {
		for (i in 0 until size) {
			if (ItemStack.isSameItemSameTags(this[i], other)) {
				return true
			}
		}

		return false
	}

	final override fun getItem(slot: Int): ItemStack {
		val item = slots[slot]

		if (item.isEmpty)
			return ItemStack.EMPTY
		else
			return item
	}

	final override fun removeItem(slot: Int, amount: Int): ItemStack {
		if (amount <= 0 || slot < 0 || slot >= size || slots[slot].isEmpty)
			return ItemStack.EMPTY

		val old = slots[slot].copy()
		val split = slots[slot].split(amount)
		trackedSlots[slot] = slots[slot].copy()
		setChanged(slot, if (slots[slot].isEmpty) ItemStack.EMPTY else slots[slot], old)

		return split
	}

	final override fun removeItemNoUpdate(slot: Int): ItemStack {
		val old = slots[slot]
		slots[slot] = ItemStack.EMPTY
		trackedSlots[slot] = ItemStack.EMPTY
		return old
	}

	final override fun setItem(slot: Int, stack: ItemStack) {
		if (slots[slot].isEmpty && stack.isEmpty || stack === slots[slot])
			return

		val old = slots[slot]
		slots[slot] = stack
		trackedSlots[slot] = stack.copy()
		setChanged(slot, stack, old)
	}

	final override fun setChanged() {
		for (slot in 0 until size) {
			setChanged(slot)
		}
	}

	fun setChanged(slot: Int) {
		if (!ItemStack.isSameItemSameTags(slots[slot], trackedSlots[slot]) || slots[slot].count != trackedSlots[slot].count) {
			setChanged(slot, slots[slot], trackedSlots[slot])
			trackedSlots[slot] = slots[slot].copy()
			// mojang соси))0)0))0)))))0)
		}
	}

	override fun stillValid(player: Player): Boolean {
		return true
	}

	final override fun clearContent() {
		for (slot in 0 until size) {
			if (!slots[slot].isEmpty) {
				val old = slots[slot]
				slots[slot] = ItemStack.EMPTY
				setChanged(slot, ItemStack.EMPTY, old)
			}
		}

		Arrays.fill(trackedSlots, ItemStack.EMPTY)
	}

	final override fun iterator(): MutableIterator<ItemStack> {
		return ContainerIterator(this)
	}
}
